-------------------------------------------------------------------------------
--
-- Modul, welches eine Instanz eines Graceful Problems behandelt
--
-- Einfach ausgedrckt schaut das Modul, ob ein gegebener Graph mit dem
-- gegebenen Labeling graceful (schn, hbsch) gelabelt ist.
--
--
-- Autor: Alexander Kiel
-- Version: 11.05.2002
-------------------------------------------------------------------------------

module Graceful.Graceful
	( Graceful
	, module Graceful.Labeling
	, validiere
	, verifiziere
	, getInstanz
	, getBeweis
	) 
    where

import Graceful.Labeling
import Graph.Type
import Graph.Util
import Graph.Viz
import Challenger
import ToDoc
import Set
import Sort
import Maybe
import FiniteMap

data Graceful = Graceful deriving Show


-- Instanzierung der Klasse Problem (problem, instanz, beweis)
--   eine Instanz muss hier ein (Graph a) sein, also ein Graph mit beliebigem
--     Knotentyp
--   ein Beweis muss hier ein (Labeling a) sein, also eine Abbildung von dem
--     beliebigem Knotentyp a auf  

instance
	( Ord knoten, Show knoten, ToDoc knoten
	, ToDoc (Graph knoten), Show (Graph knoten), Read (Graph knoten)
	, ToDoc (Labeling knoten), Show (Labeling knoten), Read (Labeling knoten)
	)
	=> Problem Graceful (Graph knoten) (Labeling knoten) where


	validiere Graceful graph labeling =
		testeAlles (text "Der Graph ist valid.") (
			[ kantenPassenZuKnotenTest $ kantenPassenZuKnoten graph
			, ( isZusammen graph
			  , text "Der Graph ist nicht zusammenhngend."
			  )
			])
    
	verifiziere Graceful graph labeling =
		testeAlles (text "Die Lsung ist richtig.") (
			[ isAnzKnotenEqLengthLabeling graph labeling
			, isSmallestLabelNull labeling
			, isDiffListTight $ sort $ getDiffList graph labeling
			])

	getInstanz Graceful graph labeling dateiName =
		-- mehr muss hier keiner machen, der Graphviz nutzt
		getGraphviz graph instanzTrans dateiName
	
	getBeweis Graceful graph labeling dateiName =
		getGraphviz graph (getBeweisTrans labeling) dateiName
		
-------------------------------------------------------------------------------
-- hier folgen Transformationen
-------------------------------------------------------------------------------

-- ganz einfache Transformation
instanzTrans :: Show knoten => GVTrans knoten
instanzTrans = GVTrans
	{ getGVProg = Default
	, getGVFormat = "gif"
	, isGVDirected = True
	, getGVNID = show		-- Graphviz Knoten ID ist einfach show knoten
	, getGVNName = show		-- Knotenname auch
	, getGVNLabel = Nothing
	, getGVNColor = Nothing
	, getGVNXAtts = Nothing
	, getGVELabel = Nothing
	, getGVEXAtts = Nothing
	}

-- um Label erweiterte Transformation
getBeweisTrans :: (Show knoten, Ord knoten)
	=> (Labeling knoten) -> (GVTrans knoten)
getBeweisTrans labeling = GVTrans
	{ getGVProg = Dot
	, getGVFormat = "gif"
	, isGVDirected = True
	, getGVNID = show
	, getGVNName = show
	, getGVNLabel = Just (getNLabel labeling)
	, getGVNColor = Nothing
	, getGVNXAtts = Nothing
	, getGVELabel = Just (getELabel labeling)
	, getGVEXAtts = Nothing
	}

-- Knotenlabel ist Graceful spezifisch
getNLabel :: Ord a => (Labeling a) -> a -> GVLabel
getNLabel labeling knoten = show (getLabel labeling knoten)

-- Kantenlabel ist Graceful spezifisch
getELabel :: Ord a => (Labeling a) -> (Kante a) -> GVLabel
getELabel labeling kante = show (getDiff labeling kante)

-------------------------------------------------------------------------------
-- hier folgen die einzelnen Tests
-------------------------------------------------------------------------------

kantenPassenZuKnotenTest :: ToDoc knoten => (Bool, Maybe knoten) -> (Bool, Doc)
kantenPassenZuKnotenTest (bool, Nothing) = (bool, text "")
kantenPassenZuKnotenTest (bool, Just knoten) =
	( bool
	, text "Es gibt Kanten, die zu dem Knoten" <+>
      ( toDoc $ fromJust (Just knoten) ) <+>
	  text "zeigen, den es im Graph nicht gibt."
    ) 

isAnzKnotenEqLengthLabeling :: (Graph a) -> (Labeling a) -> (Bool, Doc)
isAnzKnotenEqLengthLabeling graph labeling =
    ( (anzKnoten graph == sizeL labeling)
    , text "Anzahl der Knoten stimmt nicht mit der Anzahl der Labels berein."
    )

-- die Labels mssen mit der Zahl 0 anfangen
isSmallestLabelNull :: (Labeling a) -> (Bool, Doc)
isSmallestLabelNull labeling =
    ( (==) 0 $ head $ getSortedLabelList labeling
    , text "Das erste Label muss 0 sein."
    )

isDiffListTight :: [Integer] -> (Bool, Doc)
isDiffListTight diffList =
    ( isListTight diffList 0
    , text "Die Menge der Labels ist nicht dicht gegenber den natrlichen Zahlen."
    )

isLabelingGraceful :: (Graph a) -> (Labeling a) -> (Bool, Doc)
isLabelingGraceful graph labeling =
    ( False
    , text ""
    )

-------------------------------------------------------------------------------
-- hier folgen Hilfsfunktionen
-------------------------------------------------------------------------------

-- gibt eine Liste mit den Differenzen, welche an den Kanten stehen zurck
-- die Elemente der Liste sind nicht geordnet
getDiffList :: (Ord knoten) => (Graph knoten) -> (Labeling knoten) -> [Integer]
getDiffList graph labeling =
	[ abs((getLabel labeling (von kante)) - (getLabel labeling (nach kante)))
	| kante <- setToList(kanten graph)
	]

-- fhrt alle Tests durch und gibt entweder:
--   die Ausgabe (Bool, Doc) des ersten nicht bestanden Tests zurck
--   oder die positiv Meldung in der Form (True, Doc) zurck
testeAlles :: Doc -> [(Bool, Doc)] -> (Bool, Doc)
testeAlles positivDoc tests =
	getFirstError positivDoc [(b, doc) | (b, doc) <- tests, b == False]
    
getFirstError :: Doc -> [(Bool, Doc)] -> (Bool, Doc)
getFirstError positivDoc [] = (True, positivDoc)
getFirstError positivDoc es = head es

-- prft, ob eine Liste gegenber den natrlichen Zahlen ab einem bestimmten
-- Anfangswert dich ist
-- ist die Liste muss sortiert sein!
isListTight :: [Integer] -> Integer -> Bool
isListTight [] _ = True
isListTight (x:xs) y = if x == y then isListTight xs (y+1) else False
